from .publisher import Publisher
from .just import Just
from .subject import PassThroughSubject
from .url_request import URLRequestPublisher
from .async_publisher import AsyncPublisher, AsyncMapPublisher



from __future__ import annotations

from typing import Optional, Union, Callable, Any

import discord
from discord import ui

from .item import Item
from .custom import CustomButton
from .modal import Modal


class LinkButton(Item):
    def __init__(self, url: str, label: str):
        self.url = url
        self.label = label

    def to_discord_item(self, row: Optional[int]) -> ui.Item:
        button = ui.Button(style=discord.ButtonStyle.link, label=self.label, url=self.url)
        button.row = row
        return button


class Button(Item):
    def __init__(
            self,
            label: str = "",
            style: discord.ButtonStyle = discord.ButtonStyle.primary,
            disabled: bool = False,
            emoji: Optional[Union[str, discord.PartialEmoji]] = None,
            custom_id: Optional[str] = None,
    ):
        self._style: discord.ButtonStyle = style
        self._label: str = label
        self._disabled: bool = disabled
        self._emoji: Optional[Union[str, discord.PartialEmoji]] = emoji
        self._custom_id: Optional[str] = custom_id

        self.callback_func: Optional[Callable[[discord.Interaction], Any]] = None
        self.check_func: Optional[Callable[[discord.Interaction], bool]] = None
        self.modal_submit: Optional[Modal] = None

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Button):
            return NotImplemented

        return self._label == other._label\
            and self._style == other._style\
            and self._disabled == other._disabled\
            and self._emoji == other._emoji\
            and self._custom_id == other._custom_id

    def style(self, style: discord.ButtonStyle) -> Button:
        self._style = style
        return self

    def label(self, label: str) -> Button:
        self._label = label
        return self

    def disabled(self, disabled: bool = False) -> Button:
        self._disabled = disabled
        return self

    def emoji(self, emoji: Union[str, discord.PartialEmoji]) -> Button:
        self._emoji = emoji
        return self

    def custom_id(self, custom_id: str) -> Button:
        self._custom_id = custom_id
        return self

    def on_click(self, func: Callable[[discord.Interaction], Any]) -> Button:
        self.callback_func = func
        return self

    def modal(self, modal: Modal) -> Button:
        self.modal_submit = modal
        return self

    def check(self, func: Callable[[discord.Interaction], bool]) -> Button:
        self.check_func = func
        return self

    def to_discord_item(self, row: Optional[int]) -> ui.Item:
        button = CustomButton(
            self._label,
            self._style,
            self._disabled,
            self._emoji,
            self._custom_id,
            modal_submit=self.modal_submit
        )
        button.check_func = self.check_func
        button.callback_func = self.callback_func
        button.row = row
        return button




from __future__ import annotations
from typing import Optional, Callable, Union

import discord
from discord import ui

from .item import Item
from .custom import CustomSelect


class Select(Item):
    def __init__(
            self,
            placeholder: Optional[str] = None,
            min_values: int = 1,
            max_values: int = 1,
            options: Optional[list] = None,
            custom_id: Optional[str] = None,
    ) -> None:
        self._placeholder: Optional[str] = placeholder
        self._min_values: int = min_values
        self._max_values: int = max_values
        self._options: list = [] if options is None else options
        self._row: Optional[int] = None
        self._custom_id: Optional[str] = custom_id

        self.func: Optional[Callable] = None
        self.check_func: Optional[Callable[[discord.Interaction], bool]] = None

    def placeholder(self, placeholder: str) -> 'Select':
        self._placeholder = placeholder
        return self

    def min_values(self, min_values: int) -> 'Select':
        self._min_values = min_values
        return self

    def max_values(self, max_values: int) -> 'Select':
        self._max_values = max_values
        return self

    def options(self, options: list[SelectOption]) -> 'Select':
        self._options = [op.to_discord_select_option() for op in options]
        return self

    def row(self, row: int) -> 'Select':
        self._row = row
        return self

    def on_select(self, func: Callable) -> 'Select':
        self.func = func

        return self

    def custom_id(self, custom_id: str) -> 'Select':
        self._custom_id = custom_id
        return self

    def check(self, func: Callable[[discord.Interaction], bool]) -> 'Select':
        self.check_func = func
        return self

    def to_discord_item(self, row: Optional[int]) -> ui.Item:
        return CustomSelect(
            custom_id=self._custom_id,
            placeholder=self._placeholder,
            min_values=self._min_values,
            max_values=self._max_values,
            options=self._options,
            row=row,
            callback=self.func,
            check_func=self.check_func,
        )


class SelectOption:
    def __init__(
            self,
            label: str,
            value: Optional[str] = None,
            description: Optional[str] = None,
            emoji: Optional[Union[str, discord.PartialEmoji]] = None,
            default: bool = False
    ) -> None:
        self._label: str = label
        self._value: Optional[str] = label if value is None else value
        self._description: Optional[str] = description

        if isinstance(emoji, str):
            emoji = discord.PartialEmoji.from_str(emoji)

        self._emoji: Optional[str] = emoji
        self._default: bool = default

    def label(self, label: str) -> 'SelectOption':
        """Set the label of the option.
        The label of the option. This is displayed to users.
        Can only be up to 25 characters.
        This function returns the class instance to allow for fluent-style
        chaining.
        Parameters
        -----------
        label: :class:`str`
            The label to set.
        """
        self._label = label
        return self

    def value(self, value: str) -> 'SelectOption':
        """Set the value of the option.
        The value of the option. This is not displayed to users.
        If not provided when constructed then it defaults to the
        label. Can only be up to 100 characters.
        This function returns the class instance to allow for fluent-style
        chaining.
        Parameters
        -----------
        value: :class:`str`
            The value to set.
        """
        self._value = value
        return self

    def description(self, description: str) -> 'SelectOption':
        """Set the description of the option.
        An additional description of the option.
        Can only be up to 50 characters.
        This function returns the class instance to allow for fluent-style
        chaining.
        Parameters
        -----------
        description: :class:`str`
            The description to set.
        """
        self._description = description
        return self

    def emoji(self, emoji: Optional[Union[str, discord.PartialEmoji]]) -> 'SelectOption':
        """Set the emoji of the option.
        This function returns the class instance to allow for fluent-style
        chaining.
        Parameters
        -----------
        emoji: Optional[Union[:class:`str`, :class:`discord.PartialEmoji`]]
            The description to set.
        """
        if isinstance(emoji, str):
            emoji = discord.PartialEmoji.from_str(emoji)

        self._emoji = emoji
        return self

    def default(self, default: bool) -> 'SelectOption':
        """Set the default of the option.
        Whether this option is selected by default.
        This function returns the class instance to allow for fluent-style
        chaining.
        Parameters
        -----------
        default: :class:`bool`
            The default value to set.
        """
        self._default = default
        return self

    def to_discord_select_option(self) -> discord.SelectOption:
        return discord.SelectOption(
            label=self._label,
            value=self._value,
            description=self._description,
            emoji=self._emoji,
            default=self._default
        )



# flake8: noqa
from .view import View
from .tracker import ViewTracker
from .provider import MessageProvider, InteractionProvider
from .button import LinkButton, Button
from .message import Message
from .observable_object import ObservableObject
from .state import state
from .published import published
from .select import SelectOption, Select
from .page import PaginationView, PaginationButtons, PageView
from .alert import Alert, ActionButton
from .modal import Modal


__title__ = 'discord.ext.ui'
__author__ = 'sizumita'
__license__ = 'MIT'
__copyright__ = 'Copyright 2020-present sizumita'
__version__ = "3.1.8"



from __future__ import annotations
from .view import View


def state(name: str):
    def getter(instance):
        return instance.__dict__[name]

    def setter(instance, value):
        instance.__dict__[name] = value
        if isinstance(instance, View):
            instance.update_sync()

    return property(getter, setter)



from __future__ import annotations
from typing import Any, TypeVar

from .observable_object import ObservableObject

T = TypeVar('T')


def published(name: str):
    def getter(instance: T) -> Any:
        return instance.__dict__[name]

    def setter(instance: T, value: Any) -> None:
        instance.__dict__[name] = value
        if isinstance(instance, ObservableObject):
            instance.notify()

    return property(getter, setter)



from __future__ import annotations

import asyncio
from typing import Optional, TYPE_CHECKING, Any

from .message import Message
from .button import LinkButton
from .observable_object import ObservableObject

if TYPE_CHECKING:
    from .tracker import ViewTracker


class View:
    def __init__(self, loop: Optional[asyncio.AbstractEventLoop] = None):
        self._tracker: Optional['ViewTracker'] = None
        self.loop = loop or asyncio.get_event_loop()
        self._super_view: Optional[View] = None

    async def body(self) -> Message | View:
        return Message()\
            .content("Hello World!\n\ncreated by discord-ext-ui from @sizumita")\
            .item(LinkButton("https://twitter.com/sizumita", "Twitter @sizumita"))\
            .item(LinkButton("https://github.com/sizumita/discord-ext-ui", "Github discord-ext-ui"))

    async def on_appear(self) -> None:
        """
        Viewが送信された際に実行されます。
        """
        pass

    async def on_disappear(self) -> None:
        """
        Viewがstopされた際に送信されます。
        """
        pass

    async def on_update(self) -> None:
        """
        Viewが更新された際に送信されます。
        """
        pass

    def stop(self):
        self._tracker.stop()
        self.loop.create_task(self.on_disappear())

    def update_sync(self):
        if self._tracker is not None:
            self.loop.create_task(self._tracker.update())
        if self._super_view is not None:
            self._super_view.update_sync()

    def __setattr__(self, key: str, value: Any) -> None:
        if isinstance(value, ObservableObject):
            if value.view is None:
                value.view = self

        object.__setattr__(self, key, value)



from discord import ui
import discord

from typing import List, Callable, Any
from .utils import _call_any


class Modal(ui.Modal):
    def __init__(self, title: str, components: List[ui.TextInput]):
        super().__init__(title=title)
        for component in components:
            self.add_item(component)

        self._hook = None

    def hook(self, func: Callable[[discord.Interaction], Any]) -> Modal:
        self._hook = func
        return self

    async def on_submit(self, interaction: discord.Interaction) -> None:
        if self._hook is not None:
            await _call_any(self._hook, interaction)
        if not interaction.response.is_done():
            await interaction.response.defer()



from __future__ import annotations
from typing import Optional

import discord
from discord import ui


class BaseProvider:
    async def send_message(self, content: Optional[str], embeds: list[discord.Embed], view: ui.View) -> discord.Message:
        pass

    async def edit_message(self, content: Optional[str], embeds: list[discord.Embed], view: ui.View) -> discord.Message:
        pass

    def update_interaction(self, interaction: discord.Interaction):
        pass


class MessageProvider(BaseProvider):
    def __init__(self, channel: discord.TextChannel) -> None:
        self.channel = channel
        self.message: Optional[discord.Message] = None

    async def send_message(self, content: Optional[str], embeds: list[discord.Embed], view: ui.View) -> discord.Message:
        self.message = await self.channel.send(content, embeds=embeds, view=view)
        return self.message

    async def edit_message(self, content: Optional[str], embeds: list[discord.Embed], view: ui.View) -> discord.Message:
        await self.message.edit(content=content, embeds=embeds, view=view)
        return self.message


class InteractionProvider(BaseProvider):
    def __init__(self, interaction: discord.Interaction, *args, **kwargs) -> None:
        self.interaction = interaction
        self.message: Optional[discord.Message] = None
        self._args = args
        self._kwargs = kwargs

    async def send_message(self, content: Optional[str], embeds: list[discord.Embed], view: ui.View) -> discord.Message:
        resp: discord.InteractionResponse = self.interaction.response
        if resp._responded:
            followup: discord.Webhook = self.interaction.followup
            self.message = await followup.send(content, embeds=embeds, view=view, wait=True, *self._args, **self._kwargs)
        else:
            await resp.send_message(content, embeds=embeds, view=view, *self._args, **self._kwargs)
            self.message = await self.interaction.original_message()
        return self.message

    async def edit_message(self, content: Optional[str], embeds: list[discord.Embed], view: ui.View) -> discord.InteractionMessage:
        await self.interaction.edit_original_message(content=content, embeds=embeds, view=view)
        self.message = await self.interaction.original_message()
        return self.message

    def update_interaction(self, interaction: discord.Interaction):
        self.interaction = interaction
